#include <apd.h>
#include <apd_internal.h>

// See apd.h for more on this
#if APD_USE_INTERNAL_DEBUG_OBJECT == 1
    struct apd_object_struct *apd_object = NULL;
#endif

// Used by the Valgrind stack trace feature:
int _apd_uninitialized_source;
int _apd_uninitialized_target;

// ==============================================================
// Initialize the global variables in the library when the library
// is first loaded.  This uses a non-standard gcc extension
// (the __attribute__((constructor)) part).  If this is ported to
// new compilers, you should use your compiler's method for
// calling a library initializer.
// ==============================================================
#ifdef __GNUC__
static void __attribute__((constructor)) _apd_init(void)
{
// Stop stdout buffering so that the stderr messages 
// show up interspersed properly with stdout.
#ifdef SETVBUF_REVERSED /* special version for wacked systems */
setvbuf(stdout, _IONBF, NULL, 0);
#else /* Normal version for GNU systems */
setvbuf(stdout, NULL, _IONBF, 0);
#endif

// Set adp_object to default
apd_object = apd_initialize(
    APD_DEFAULT_REPORT_THRESHOLD,  /* int report_threshold */
    APD_DEFAULT_ASSERT_THRESHOLD,  /* int assert_threshold */
    APD_DEFAULT_MAX_FUNCTION_DEPTH,  /* int max_function_depth */
    APD_DEFAULT_REPORT_THRESHOLD_DECREASES_WITH_STACK,
    APD_DEFAULT_ERROR_STREAM  /* FILE *error_stream) */
    );
}

static void __attribute__((destructor)) _apd_end(void)
{
// Free memory allocated in apd_initialize if internal apd_object
#if APD_USE_INTERNAL_DEBUG_OBJECT == 1
    apd_end(&apd_object);
#endif
}
#endif /* __GNUC__ */

// ===============================================
// ===============================================
struct apd_object_struct *
apd_initialize(
    int report_threshold,
    int assert_threshold,
    uint max_function_depth,
    bool report_threshold_decreases_with_stack,
    FILE *error_stream)
{
struct apd_object_struct *debug_object_return = NULL;
uint level, index;

// Check error_stream now since we'll use it to report internal_debug info
if (error_stream == NULL) { error_stream = stdout; }

// APD_INTERNAL_REPORT uses apd_object, so until that is set up with the correct error stream,
// we'll just use APD_INTERNAL_DEBUG_COMPILE to decide whether or not to display debug output
#if APD_INTERNAL_DEBUG_COMPILE == 1
fprintf(error_stream, "apd_initialize(): starting\n");
fprintf(error_stream, "changing report threshold level to %d\n", report_threshold);
fprintf(error_stream, "changing assert threshold level to %d\n", assert_threshold);
fprintf(error_stream, "changing max_function_depth to %d\n", max_function_depth);
fprintf(error_stream, "changing report_threshold_decreases_with_stack to %d\n", report_threshold_decreases_with_stack);
fprintf(error_stream, "changing file to stream at $%lX\n", (unsigned long int)error_stream);
#endif

// Allocate debug_object_return
debug_object_return = (struct apd_object_struct *)malloc(sizeof(struct apd_object_struct));
#if APD_INTERNAL_DEBUG_COMPILE == 1
fprintf(error_stream, "allocated apd_object at $%lX\n", (unsigned long int)debug_object_return);
#endif
if (debug_object_return == NULL) APD_DIE("error allocating apd_object");
// Allocate backup copy
debug_object_return->backup = (struct apd_object_struct *)malloc(sizeof(struct apd_object_struct));
#if APD_INTERNAL_DEBUG_COMPILE == 1
fprintf(error_stream, "allocated apd_object->backup at $%lX\n", (unsigned long int)debug_object_return->backup);
#endif
if (debug_object_return->backup == NULL) APD_DIE("error allocating apd_object->backup");

// Empty description stacks
debug_object_return->description_stack = NULL;
debug_object_return->description_stack_size = 0;

// Now set everything normally
if (report_threshold < 0)
  {
    _APD_DIE_GUTS(APD_DEFAULT_ERROR_STREAM, __FILE__, __LINE__, APD_FUNCTION_NAME,
          "report_threshold must be greater or equalt to 0 (report_threshold = %d)", report_threshold);
  }
debug_object_return->report_threshold = report_threshold;

if (assert_threshold < 0)
  {
    _APD_DIE_GUTS(APD_DEFAULT_ERROR_STREAM, __FILE__, __LINE__, APD_FUNCTION_NAME,
          "assert_threshold must be greater or equalt to 0 (assert_threshold = %d)", assert_threshold);
  }
debug_object_return->assert_threshold = assert_threshold;

if (max_function_depth < 0)
  {
    _APD_DIE_GUTS(APD_DEFAULT_ERROR_STREAM, __FILE__, __LINE__, APD_FUNCTION_NAME,
          "max_function_depth must be greater or equalt to 0 (max_function_depth = %d)", max_function_depth);
  }
debug_object_return->max_function_depth = max_function_depth;

if (report_threshold_decreases_with_stack != 0 && report_threshold_decreases_with_stack != 1)
  {
    _APD_DIE_GUTS(APD_DEFAULT_ERROR_STREAM, __FILE__, __LINE__, APD_FUNCTION_NAME,
          "report_threshold_decreases_with_stack must be 0 or 1 (report_threshold_decreases_with_stack = %d)",
          report_threshold_decreases_with_stack);
  }
debug_object_return->report_threshold_decreases_with_stack = report_threshold_decreases_with_stack;

// error_stream was already checked above
debug_object_return->error_stream = error_stream;

// Set up level_spacers
debug_object_return->level_spacers = (char **)malloc(sizeof(char *) * max_function_depth);
for (level = 0; level < max_function_depth; level++)
  {
    debug_object_return->level_spacers[level] = (char *)malloc(sizeof(char) * (level+2));
    for (index = 0; index < (level+1); index++)
      {
        debug_object_return->level_spacers[level][index] = APD_LEVEL_SPACER_CHARACTER;
      }
    debug_object_return->level_spacers[level][level+1] = '\0';  // end of string
  }

if (debug_object_return->report_threshold >= 3)
  {
    fprintf(debug_object_return->error_stream, "apd_initialize(): initialized\n");
    fflush(debug_object_return->error_stream);
  }

// Default apd debugging and internal consistency checking to off
debug_object_return->internal_debug = 0;
debug_object_return->check_internal = 0;
debug_object_return->internal_debug_level = APD_DEFAULT_INTERNAL_DEBUG_LEVEL;  // Defined in apd_internal.h
debug_object_return->action_extra_info = 0;

// Default valgrind stack trace to on
debug_object_return->valgrind_stack_trace = 1;

// No function currently on the function description stack
debug_object_return->description_stack = NULL;
debug_object_return->description_stack_size = 0;
debug_object_return->current_function_name = NULL;
debug_object_return->backup->description_stack = NULL;
debug_object_return->backup->description_stack_size = 0;
debug_object_return->backup->current_function_name = NULL;

// Refresh apd_object's backup copy since we just changed it
APD_BACKUP_DEBUG_OBJECT(debug_object_return);

#if APD_INTERNAL_DEBUG_COMPILE == 1
fprintf(debug_object_return->error_stream, "apd_initialize(): ending\n");
#endif
return debug_object_return;
} // end apd_initialize()

// ===============================================
// Deallocate the apd data structures
// ===============================================
void apd_end(struct apd_object_struct **apd_object_internal)
{
uint level;

if (*apd_object_internal == NULL)
  {
    return;
  }

printf("apd_object_internal = %lX\n", (unsigned long int)*apd_object_internal);
APD_INTERNAL_REPORT(2, "deallocating apd data structures");

// Free level_spacers
for (level = 0; level < (*apd_object_internal)->max_function_depth; level++)
  {
    free((*apd_object_internal)->level_spacers[level]);
  }
free((*apd_object_internal)->level_spacers);

// Deallocate backup copy
free((*apd_object_internal)->backup);

// Deallocate apd_object_internal
free(*apd_object_internal);

*apd_object_internal = NULL;
} // end apd_end()

// ========================================
// ========================================
void apd_set(struct apd_object_struct *apd_object_internal, int set_type, int value)
{
APD_INTERNAL_REPORT(5, "start");
_apd_check_if_function_tracking_initialized(apd_object_internal);

switch(set_type)
  {
    case APD_SET_REPORT_THRESHOLD:
        APD_INTERNAL_ASSERT(value >= 0, "");
        apd_object_internal->report_threshold = value;
        break;
        
    case APD_SET_ASSERT_THRESHOLD:
        APD_INTERNAL_ASSERT(value >= 0, "");
        apd_object_internal->assert_threshold = value;
        break;

    case APD_SET_MAX_FUNCTION_DEPTH:
        APD_INTERNAL_ASSERT(value >= 0, "");
        apd_object_internal->max_function_depth = value;
        break;

    case APD_SET_INTERNAL_DEBUG:
        APD_INTERNAL_ASSERT((value == 0) || (value == 1), "");
        apd_object_internal->internal_debug = value;
        if (value == 0)
          {
            APD_INTERNAL_REPORT(1, "internal_debug turned off");
          }
        else
          {
            APD_INTERNAL_REPORT(1, "internal_debug turned on");
          }
        break;

    case APD_SET_CHECK_INTERNAL:
        APD_INTERNAL_ASSERT((value == 0) || (value == 1), "");
        apd_object_internal->check_internal = value;
        break;

    case APD_SET_ACTION_EXTRA_INFO_ON:
        APD_INTERNAL_ASSERT((value == 0) || (value == 1), "");
        apd_object_internal->action_extra_info = value;
        break;

    case APD_SET_REPORT_THRESHOLD_DECREASES_WITH_STACK:
        APD_INTERNAL_ASSERT((value == 0) || (value == 1), "");
        apd_object_internal->report_threshold_decreases_with_stack = value;
        break;

    case APD_SET_VALGRIND_STACK_TRACE:
        APD_INTERNAL_ASSERT((value == 0) || (value == 1), "");
        apd_object_internal->valgrind_stack_trace = value;
        break;

    default:
        APD_DIE("unknown set_type");
        break;
  } // end switch(set_type)

// Refresh apd_object's backup copy since we just changed it
APD_BACKUP_DEBUG_OBJECT(apd_object_internal);

APD_INTERNAL_REPORT(5, "end");
} // end apd_set()


// ========================================
// ========================================
void apd_set_file(struct apd_object_struct *apd_object_internal, int set_type, FILE *value)
{
APD_INTERNAL_REPORT(5, "start");
_apd_check_if_function_tracking_initialized(apd_object_internal);

switch(set_type)
  {
    case APD_SET_ERROR_STREAM:
        APD_INTERNAL_ASSERT(value != NULL, "");
        apd_object_internal->error_stream = value;
        break;

    default:
        APD_DIE("unknown set_type");
        break;
  } // end switch(set_type)

// Refresh apd_object's backup copy since we just changed it
APD_BACKUP_DEBUG_OBJECT(apd_object_internal);

APD_INTERNAL_REPORT(5, "end");
} // end apd_set_file()


// ++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++
// This function makes sure you don't try to use the debug system
// without initializing it (which is very bad).  It only works if you follow
// the advice of the declaration template for apd_initialize()
// and initialize your apd_object to NULL, or if you use the built-in
// global debug object (if APD_USE_INTERNAL_DEBUG_OBJECT
// is #define'ed)
// ++++++++++++++++++++++++++++++++++++++++++++++
inline
void _apd_check_if_function_tracking_initialized(struct apd_object_struct *apd_object_internal)
{
APD_INTERNAL_REPORT(5, "start");
if (apd_object_internal == NULL)
  {
    fprintf(APD_DEFAULT_ERROR_STREAM, "Fatal error: function tracking not initialized, but "
                    "a function tracking routine was called.\n");
    exit(EXIT_FAILURE);
  }
APD_INTERNAL_REPORT(5, "end");
} // end _apd_check_if_function_tracking_initialized()
